@@ -4,7 +4,7 @@
 
    Author: Michal Zalewski <lcamtuf@google.com>
 
-   Copyright 2009, 2010 by Google Inc. All Rights Reserved.
+   Copyright 2009, 2010, 2011 by Google Inc. All Rights Reserved.
 
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
@@ -133,8 +133,10 @@ struct pivot_desc {
 
   /* Injection attack logic scratchpad: */
 
-  struct http_request*  misc_req[10];           /* Saved requests            */
-  struct http_response* misc_res[10];           /* Saved responses           */
+#define MISC_ENTRIES 10
+
+  struct http_request*  misc_req[MISC_ENTRIES]; /* Saved requests            */
+  struct http_response* misc_res[MISC_ENTRIES]; /* Saved responses           */
   u8 misc_cnt;                                  /* Request / response count  */
 
   u8 i_skip[15];                                /* Injection step skip flags */
@@ -262,6 +264,7 @@ u8 is_c_sens(struct pivot_desc* pv);
 
 #define PROB_CACHE_LOW          30701           /* Cache nit-picking         */
 
+#define PROB_PROLOGUE           30801           /* User-supplied prologue    */
 
 /* - Moderate severity issues (data compromise): */
 
@@ -375,19 +378,19 @@ void wordlist_confirm_word(u8* text);
 
 /* Returns wordlist item at a specified offset (NULL if no more available). */
 
-u8* wordlist_get_word(u32 offset);
+u8* wordlist_get_word(u32 offset, u8* specific);
 
 /* Returns keyword candidate at a specified offset (or NULL). */
 
-u8* wordlist_get_guess(u32 offset);
+u8* wordlist_get_guess(u32 offset, u8* specific);
 
 /* Returns extension at a specified offset (or NULL). */
 
-u8* wordlist_get_extension(u32 offset);
+u8* wordlist_get_extension(u32 offset, u8 specific);
 
 /* Loads keywords from file. */
 
-void load_keywords(u8* fname, u32 purge_age);
+void load_keywords(u8* fname, u8 read_only, u32 purge_age);
 
 /* Saves all keywords to a file. */
 
